module Hence
  module Assembler
    extend self

    def assemble(ast : Parser::AST, opcode_names : Hash(String, UInt8), data = Data.new) : Bytes
      ast.body.each do |node|
        data.constants["OFFSET"] = Parser::AST::NumberArg.new(data.offset)

        case node
        when Parser::AST::ConstantDeclarationNode
          raise "Constants can't be redefined" if data.constants[node.name]?

          data.constants[node.name] = node.value
        when Parser::AST::LabelNode
          raise "Labels can't be redefined" if data.constants[node.name]?

          data.constants[node.name] = Parser::AST::NumberArg.new(data.offset)
        when Parser::AST::MacroCallNode
          MACROS[node.name].call(data, node.args)
        when Parser::AST::CallNode
          data.bin[data.offset] = opcode_names[node.name]
          arg = node.arg.try(&.resolve(data)) || Bytes.new(2)
          data.bin[data.offset + 1] = arg[0]
          data.bin[data.offset + 2] = arg[1]

          data.offset += 3
        else
          raise "Unexpected node type: #{typeof(node)}"
        end

        raise "Offset out of bounds: 0x#{data.offset.to_s(16).rjust(4, '0')} > 0x#{data.bin.size.to_s(16).rjust(4, '0')}" if data.offset > data.bin.size
      end

      data.bin
    end

    class Data
      property bin
      property offset
      property constants

      def initialize(
        @bin = Bytes.new({{ 32 * 1024 }}),
        @offset = 0_u16,
        @constants = {} of String => Parser::AST::Arg
      )
      end
    end

    MACROS = {
      "org" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        data.offset = args[0].resolve_as_number(data)
      end,
      "org_add" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        data.offset += args[0].resolve_as_number(data)
      end,
      "org_sub" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        data.offset -= args[0].resolve_as_number(data)
      end,
      "bytes" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        args.each do |a|
          a.resolve(data).each do |b|
            data.bin[data.offset] = b
            data.offset += 1
          end
        end
      end,
      "bw" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        string_arg = String.new(args[0].resolve(data)).gsub(/\\n/, '\n').to_slice
        pos_start = args[1]?.try(&.resolve_as_number(data)) || 0_u16
        pos_end = args[2]?.try(&.resolve_as_number(data)) || string_arg.size.to_u16

        string_arg[pos_start...pos_end].each do |b|
          data.bin[data.offset] = b
          data.offset += 1
        end
      end,
      "include" => ->(_data : Data, _args : Array(Parser::AST::Arg)) do
        raise "Unimplemented"
      end,
    }
  end
end
